@@ -35,12 +35,13 @@
 
 GST_DEBUG_CATEGORY_STATIC (GST_CAT_PERFORMANCE);
 
+#define GST_FFMPEG_VIDEO_CODEC_FRAME_FLAG_ALLOCATED (1<<15)
+
 #define MAX_TS_MASK 0xff
 
 #define DEFAULT_LOWRES			0
 #define DEFAULT_SKIPFRAME		0
 #define DEFAULT_DIRECT_RENDERING	TRUE
-#define DEFAULT_DEBUG_MV		FALSE
 #define DEFAULT_MAX_THREADS		0
 #define DEFAULT_OUTPUT_CORRUPT		TRUE
 #define REQUIRED_POOL_MAX_BUFFERS       32
@@ -85,7 +86,7 @@ static void gst_ffmpegviddec_get_property (GObject * object,
     guint prop_id, GValue * value, GParamSpec * pspec);
 
 static gboolean gst_ffmpegviddec_negotiate (GstFFMpegVidDec * ffmpegdec,
-    AVCodecContext * context, AVFrame * picture);
+    AVCodecContext * context, AVFrame * picture, GstBufferFlags flags);
 
 /* some sort of bufferpool handling, but different */
 static int gst_ffmpegviddec_get_buffer2 (AVCodecContext * context,
@@ -97,7 +98,7 @@ static GstFlowReturn gst_ffmpegviddec_finish (GstVideoDecoder * decoder);
 static GstFlowReturn gst_ffmpegviddec_drain (GstVideoDecoder * decoder);
 
 static gboolean picture_changed (GstFFMpegVidDec * ffmpegdec,
-    AVFrame * picture);
+    AVFrame * picture, gboolean one_field);
 static gboolean context_changed (GstFFMpegVidDec * ffmpegdec,
     AVCodecContext * context);
 
@@ -169,6 +170,22 @@ gst_ffmpegviddec_thread_type_get_type (void)
   return ffmpegdec_thread_type_type;
 }
 
+static GstCaps *
+dup_caps_with_alternate (GstCaps * caps)
+{
+  GstCaps *with_alternate;
+  GstCapsFeatures *features;
+
+  with_alternate = gst_caps_copy (caps);
+  features = gst_caps_features_new (GST_CAPS_FEATURE_FORMAT_INTERLACED, NULL);
+  gst_caps_set_features_simple (with_alternate, features);
+
+  gst_caps_set_simple (with_alternate, "interlace-mode", G_TYPE_STRING,
+      "alternate", NULL);
+
+  return with_alternate;
+}
+
 static void
 gst_ffmpegviddec_base_init (GstFFMpegVidDecClass * klass)
 {
@@ -206,6 +223,7 @@ gst_ffmpegviddec_base_init (GstFFMpegVidDecClass * klass)
     GST_DEBUG ("Couldn't get source caps for decoder '%s'", in_plugin->name);
     srccaps = gst_caps_from_string ("video/x-raw");
   }
+  gst_caps_append (srccaps, dup_caps_with_alternate (srccaps));
 
   /* pad templates */
   sinktempl = gst_pad_template_new ("sink", GST_PAD_SINK,
@@ -249,10 +267,13 @@ gst_ffmpegviddec_class_init (GstFFMpegVidDecClass * klass)
       g_param_spec_boolean ("direct-rendering", "Direct Rendering",
           "Enable direct rendering", DEFAULT_DIRECT_RENDERING,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+#ifndef GST_REMOVE_DEPRECATED
   g_object_class_install_property (gobject_class, PROP_DEBUG_MV,
       g_param_spec_boolean ("debug-mv", "Debug motion vectors",
-          "Whether libav should print motion vectors on top of the image",
-          DEFAULT_DEBUG_MV, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+          "Whether to print motion vectors on top of the image "
+          "(deprecated, non-functional)", FALSE,
+          G_PARAM_DEPRECATED | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+#endif
   g_object_class_install_property (gobject_class, PROP_OUTPUT_CORRUPT,
       g_param_spec_boolean ("output-corrupt", "Output corrupt buffers",
           "Whether libav should output frames even if corrupted",
@@ -302,7 +323,6 @@ gst_ffmpegviddec_init (GstFFMpegVidDec * ffmpegdec)
   ffmpegdec->opened = FALSE;
   ffmpegdec->skip_frame = ffmpegdec->lowres = 0;
   ffmpegdec->direct_rendering = DEFAULT_DIRECT_RENDERING;
-  ffmpegdec->debug_mv = DEFAULT_DEBUG_MV;
   ffmpegdec->max_threads = DEFAULT_MAX_THREADS;
   ffmpegdec->output_corrupt = DEFAULT_OUTPUT_CORRUPT;
   ffmpegdec->thread_type = DEFAULT_THREAD_TYPE;
@@ -320,12 +340,7 @@ gst_ffmpegviddec_finalize (GObject * object)
   GstFFMpegVidDec *ffmpegdec = (GstFFMpegVidDec *) object;
 
   av_frame_free (&ffmpegdec->picture);
-
-  if (ffmpegdec->context != NULL) {
-    gst_ffmpeg_avcodec_close (ffmpegdec->context);
-    av_free (ffmpegdec->context);
-    ffmpegdec->context = NULL;
-  }
+  avcodec_free_context (&ffmpegdec->context);
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -342,6 +357,18 @@ gst_ffmpegviddec_context_set_flags (AVCodecContext * context, guint flags,
     context->flags &= ~flags;
 }
 
+static void
+gst_ffmpegviddec_context_set_flags2 (AVCodecContext * context, guint flags,
+    gboolean enable)
+{
+  g_return_if_fail (context != NULL);
+
+  if (enable)
+    context->flags2 |= flags;
+  else
+    context->flags2 &= ~flags;
+}
+
 /* with LOCK */
 static gboolean
 gst_ffmpegviddec_close (GstFFMpegVidDec * ffmpegdec, gboolean reset)
@@ -363,13 +390,11 @@ gst_ffmpegviddec_close (GstFFMpegVidDec * ffmpegdec, gboolean reset)
 
   gst_buffer_replace (&ffmpegdec->palette, NULL);
 
-  if (ffmpegdec->context->extradata) {
-    av_free (ffmpegdec->context->extradata);
-    ffmpegdec->context->extradata = NULL;
-  }
+  av_freep (&ffmpegdec->context->extradata);
   if (reset) {
-    if (avcodec_get_context_defaults3 (ffmpegdec->context,
-            oclass->in_plugin) < 0) {
+    avcodec_free_context (&ffmpegdec->context);
+    ffmpegdec->context = avcodec_alloc_context3 (oclass->in_plugin);
+    if (ffmpegdec->context == NULL) {
       GST_DEBUG_OBJECT (ffmpegdec, "Failed to set context defaults");
       return FALSE;
     }
@@ -431,6 +456,33 @@ gst_ffmpegviddec_get_palette (GstFFMpegVidDec * ffmpegdec,
   }
 }
 
+static gboolean
+gst_ffmpegviddec_needs_reset (GstFFMpegVidDec * ffmpegdec,
+    GstVideoCodecState * state)
+{
+  GstCaps *last_caps, *new_caps;
+  gboolean needs_reset;
+
+  if (ffmpegdec->last_caps == NULL)
+    return TRUE;
+
+  last_caps = gst_caps_copy (ffmpegdec->last_caps);
+  new_caps = gst_caps_copy (state->caps);
+
+  /* Simply ignore framerate for now, this could easily be evolved per CODEC if
+   * future issue are found.*/
+  gst_structure_remove_field (gst_caps_get_structure (last_caps, 0),
+      "framerate");
+  gst_structure_remove_field (gst_caps_get_structure (new_caps, 0),
+      "framerate");
+
+  needs_reset = !gst_caps_is_equal (last_caps, new_caps);
+
+  gst_caps_unref (last_caps);
+  gst_caps_unref (new_caps);
+
+  return needs_reset;
+}
 
 static gboolean
 gst_ffmpegviddec_set_format (GstVideoDecoder * decoder,
@@ -444,18 +496,14 @@ gst_ffmpegviddec_set_format (GstVideoDecoder * decoder,
   ffmpegdec = (GstFFMpegVidDec *) decoder;
   oclass = (GstFFMpegVidDecClass *) (G_OBJECT_GET_CLASS (ffmpegdec));
 
-  if (ffmpegdec->last_caps != NULL &&
-      gst_caps_is_equal (ffmpegdec->last_caps, state->caps)) {
-    return TRUE;
-  }
-
   GST_DEBUG_OBJECT (ffmpegdec, "setcaps called");
 
   GST_OBJECT_LOCK (ffmpegdec);
-  /* stupid check for VC1 */
-  if ((oclass->in_plugin->id == AV_CODEC_ID_WMV3) ||
-      (oclass->in_plugin->id == AV_CODEC_ID_VC1))
-    oclass->in_plugin->id = gst_ffmpeg_caps_to_codecid (state->caps, NULL);
+
+  if (!gst_ffmpegviddec_needs_reset (ffmpegdec, state)) {
+    gst_caps_replace (&ffmpegdec->last_caps, state->caps);
+    goto update_state;
+  }
 
   /* close old session */
   if (ffmpegdec->opened) {
@@ -520,10 +568,6 @@ gst_ffmpegviddec_set_format (GstVideoDecoder * decoder,
   ffmpegdec->context->lowres = ffmpegdec->lowres;
   ffmpegdec->context->skip_frame = ffmpegdec->skip_frame;
 
-  /* ffmpeg can draw motion vectors on top of the image (not every decoder
-   * supports it) */
-  ffmpegdec->context->debug_mv = ffmpegdec->debug_mv;
-
   if (ffmpegdec->thread_type) {
     GST_DEBUG_OBJECT (ffmpegdec, "Use requested thread type 0x%x",
         ffmpegdec->thread_type);
@@ -559,12 +603,31 @@ gst_ffmpegviddec_set_format (GstVideoDecoder * decoder,
   } else
     ffmpegdec->context->thread_count = ffmpegdec->max_threads;
 
+  if (oclass->in_plugin->id == AV_CODEC_ID_H264) {
+    GstStructure *s = gst_caps_get_structure (state->caps, 0);
+    const char *alignment;
+    gboolean nal_aligned;
+
+    alignment = gst_structure_get_string (s, "alignment");
+    nal_aligned = !g_strcmp0 (alignment, "nal");
+    if (nal_aligned) {
+      if (ffmpegdec->context->thread_type == FF_THREAD_FRAME)
+        goto nal_only_slice;
+      ffmpegdec->context->thread_type = FF_THREAD_SLICE;
+    }
+
+    gst_ffmpegviddec_context_set_flags2 (ffmpegdec->context,
+        AV_CODEC_FLAG2_CHUNKS, nal_aligned);
+    gst_video_decoder_set_subframe_mode (decoder, nal_aligned);
+  }
+
   /* open codec - we don't select an output pix_fmt yet,
    * simply because we don't know! We only get it
    * during playback... */
   if (!gst_ffmpegviddec_open (ffmpegdec))
     goto open_failed;
 
+update_state:
   if (ffmpegdec->input_state)
     gst_video_codec_state_unref (ffmpegdec->input_state);
   ffmpegdec->input_state = gst_video_codec_state_ref (state);
@@ -598,6 +661,12 @@ open_failed:
     GST_DEBUG_OBJECT (ffmpegdec, "Failed to open");
     goto done;
   }
+nal_only_slice:
+  {
+    GST_ERROR_OBJECT (ffmpegdec,
+        "Can't do NAL aligned H.264 with frame threading.");
+    goto done;
+  }
 }
 
 typedef struct
@@ -633,6 +702,8 @@ gst_ffmpegviddec_video_frame_free (GstFFMpegVidDec * ffmpegdec,
 
   if (frame->mapped)
     gst_video_frame_unmap (&frame->vframe);
+  GST_VIDEO_CODEC_FRAME_FLAG_UNSET (frame->frame,
+      GST_FFMPEG_VIDEO_CODEC_FRAME_FLAG_ALLOCATED);
   gst_video_decoder_release_frame (GST_VIDEO_DECODER (ffmpegdec), frame->frame);
   gst_buffer_replace (&frame->buffer, NULL);
   if (frame->avbuffer) {
@@ -711,7 +782,7 @@ gst_ffmpegvideodec_prepare_dr_pool (GstFFMpegVidDec * ffmpegdec,
 
 static void
 gst_ffmpegviddec_ensure_internal_pool (GstFFMpegVidDec * ffmpegdec,
-    AVFrame * picture)
+    AVFrame * picture, GstVideoInterlaceMode interlace_mode)
 {
   GstAllocationParams params = DEFAULT_ALLOC_PARAM;
   GstVideoInfo info;
@@ -738,7 +809,13 @@ gst_ffmpegviddec_ensure_internal_pool (GstFFMpegVidDec * ffmpegdec,
     gst_pad_mark_reconfigure (GST_VIDEO_DECODER_SRC_PAD (ffmpegdec));
 
   format = gst_ffmpeg_pixfmt_to_videoformat (picture->format);
-  gst_video_info_set_format (&info, format, picture->width, picture->height);
+
+  if (interlace_mode == GST_VIDEO_INTERLACE_MODE_ALTERNATE) {
+    gst_video_info_set_interlaced_format (&info, format, interlace_mode,
+        picture->width, 2 * picture->height);
+  } else {
+    gst_video_info_set_format (&info, format, picture->width, picture->height);
+  }
 
   /* If we have not yet been negotiated, a NONE format here would
    * result in invalid initial dimension alignments, and potential
@@ -820,6 +897,8 @@ gst_ffmpegviddec_get_buffer2 (AVCodecContext * context, AVFrame * picture,
 
   /* now it has a buffer allocated, so it is real and will also
    * be _released */
+  GST_VIDEO_CODEC_FRAME_FLAG_SET (frame,
+      GST_FFMPEG_VIDEO_CODEC_FRAME_FLAG_ALLOCATED);
   GST_VIDEO_CODEC_FRAME_FLAG_UNSET (frame,
       GST_VIDEO_CODEC_FRAME_FLAG_DECODE_ONLY);
 
@@ -840,7 +919,10 @@ gst_ffmpegviddec_get_buffer2 (AVCodecContext * context, AVFrame * picture,
   if (!gst_ffmpegviddec_can_direct_render (ffmpegdec))
     goto no_dr;
 
-  gst_ffmpegviddec_ensure_internal_pool (ffmpegdec, picture);
+  gst_ffmpegviddec_ensure_internal_pool (ffmpegdec, picture,
+      GST_BUFFER_FLAG_IS_SET (frame->input_buffer,
+          GST_VIDEO_BUFFER_FLAG_ONEFIELD) ? GST_VIDEO_INTERLACE_MODE_ALTERNATE :
+      GST_VIDEO_INTERLACE_MODE_PROGRESSIVE);
 
   ret = gst_buffer_pool_acquire_buffer (ffmpegdec->internal_pool,
       &frame->output_buffer, NULL);
@@ -952,11 +1034,14 @@ no_frame:
 }
 
 static gboolean
-picture_changed (GstFFMpegVidDec * ffmpegdec, AVFrame * picture)
+picture_changed (GstFFMpegVidDec * ffmpegdec, AVFrame * picture,
+    gboolean one_field)
 {
   gint pic_field_order = 0;
 
-  if (picture->interlaced_frame) {
+  if (one_field) {
+    pic_field_order = ffmpegdec->pic_field_order;
+  } else if (picture->interlaced_frame) {
     if (picture->repeat_pict)
       pic_field_order |= GST_VIDEO_BUFFER_FLAG_RFF;
     if (picture->top_field_first)
@@ -984,7 +1069,7 @@ context_changed (GstFFMpegVidDec * ffmpegdec, AVCodecContext * context)
 
 static gboolean
 update_video_context (GstFFMpegVidDec * ffmpegdec, AVCodecContext * context,
-    AVFrame * picture)
+    AVFrame * picture, gboolean one_field)
 {
   gint pic_field_order = 0;
 
@@ -995,7 +1080,7 @@ update_video_context (GstFFMpegVidDec * ffmpegdec, AVCodecContext * context,
       pic_field_order |= GST_VIDEO_BUFFER_FLAG_TFF;
   }
 
-  if (!picture_changed (ffmpegdec, picture)
+  if (!picture_changed (ffmpegdec, picture, one_field)
       && !context_changed (ffmpegdec, context))
     return FALSE;
 
@@ -1185,7 +1270,7 @@ content_light_metadata_av_to_gst (AVContentLightMetadata * av,
 
 static gboolean
 gst_ffmpegviddec_negotiate (GstFFMpegVidDec * ffmpegdec,
-    AVCodecContext * context, AVFrame * picture)
+    AVCodecContext * context, AVFrame * picture, GstBufferFlags flags)
 {
   GstVideoFormat fmt;
   GstVideoInfo *in_info, *out_info;
@@ -1193,17 +1278,54 @@ gst_ffmpegviddec_negotiate (GstFFMpegVidDec * ffmpegdec,
   gint fps_n, fps_d;
   GstClockTime latency;
   GstStructure *in_s;
+  GstVideoInterlaceMode interlace_mode;
+  gint caps_height;
+  gboolean one_field = ! !(flags & GST_VIDEO_BUFFER_FLAG_ONEFIELD);
 
-  if (!update_video_context (ffmpegdec, context, picture))
+  if (!update_video_context (ffmpegdec, context, picture, one_field))
     return TRUE;
 
+  caps_height = ffmpegdec->pic_height;
+
   fmt = gst_ffmpeg_pixfmt_to_videoformat (ffmpegdec->pic_pix_fmt);
   if (G_UNLIKELY (fmt == GST_VIDEO_FORMAT_UNKNOWN))
     goto unknown_format;
 
-  output_state =
-      gst_video_decoder_set_output_state (GST_VIDEO_DECODER (ffmpegdec), fmt,
-      ffmpegdec->pic_width, ffmpegdec->pic_height, ffmpegdec->input_state);
+  /* set the interlaced flag */
+  in_s = gst_caps_get_structure (ffmpegdec->input_state->caps, 0);
+  if (flags & GST_VIDEO_BUFFER_FLAG_ONEFIELD) {
+    /* TODO: we don't get that information from ffmpeg, so copy it from
+     * the parser */
+    interlace_mode = GST_VIDEO_INTERLACE_MODE_ALTERNATE;
+    caps_height = 2 * caps_height;
+  } else if (!gst_structure_has_field (in_s, "interlace-mode")) {
+    if (ffmpegdec->pic_interlaced) {
+      if (ffmpegdec->pic_field_order_changed ||
+          (ffmpegdec->pic_field_order & GST_VIDEO_BUFFER_FLAG_RFF)) {
+        interlace_mode = GST_VIDEO_INTERLACE_MODE_MIXED;
+      } else {
+        interlace_mode = GST_VIDEO_INTERLACE_MODE_INTERLEAVED;
+      }
+    } else {
+      interlace_mode = GST_VIDEO_INTERLACE_MODE_PROGRESSIVE;
+    }
+  } else {
+    GstVideoInfo info;
+
+    gst_video_info_from_caps (&info, ffmpegdec->input_state->caps);
+    interlace_mode = info.interlace_mode;
+  }
+
+  if (interlace_mode == GST_VIDEO_INTERLACE_MODE_ALTERNATE)
+    output_state =
+        gst_video_decoder_set_interlaced_output_state (GST_VIDEO_DECODER
+        (ffmpegdec), fmt, interlace_mode, ffmpegdec->pic_width, caps_height,
+        ffmpegdec->input_state);
+  else
+    output_state =
+        gst_video_decoder_set_output_state (GST_VIDEO_DECODER
+        (ffmpegdec), fmt, ffmpegdec->pic_width, caps_height,
+        ffmpegdec->input_state);
   if (ffmpegdec->output_state)
     gst_video_codec_state_unref (ffmpegdec->output_state);
   ffmpegdec->output_state = output_state;
@@ -1211,26 +1333,15 @@ gst_ffmpegviddec_negotiate (GstFFMpegVidDec * ffmpegdec,
   in_info = &ffmpegdec->input_state->info;
   out_info = &ffmpegdec->output_state->info;
 
-  /* set the interlaced flag */
-  in_s = gst_caps_get_structure (ffmpegdec->input_state->caps, 0);
-
-  if (!gst_structure_has_field (in_s, "interlace-mode")) {
-    if (ffmpegdec->pic_interlaced) {
-      if (ffmpegdec->pic_field_order_changed ||
-          (ffmpegdec->pic_field_order & GST_VIDEO_BUFFER_FLAG_RFF)) {
-        out_info->interlace_mode = GST_VIDEO_INTERLACE_MODE_MIXED;
-      } else {
-        out_info->interlace_mode = GST_VIDEO_INTERLACE_MODE_INTERLEAVED;
-        if ((ffmpegdec->pic_field_order & GST_VIDEO_BUFFER_FLAG_TFF))
-          GST_VIDEO_INFO_FIELD_ORDER (out_info) =
-              GST_VIDEO_FIELD_ORDER_TOP_FIELD_FIRST;
-        else
-          GST_VIDEO_INFO_FIELD_ORDER (out_info) =
-              GST_VIDEO_FIELD_ORDER_BOTTOM_FIELD_FIRST;
-      }
-    } else {
-      out_info->interlace_mode = GST_VIDEO_INTERLACE_MODE_PROGRESSIVE;
-    }
+  out_info->interlace_mode = interlace_mode;
+  if (!gst_structure_has_field (in_s, "interlace-mode")
+      && interlace_mode == GST_VIDEO_INTERLACE_MODE_INTERLEAVED) {
+    if ((ffmpegdec->pic_field_order & GST_VIDEO_BUFFER_FLAG_TFF))
+      GST_VIDEO_INFO_FIELD_ORDER (out_info) =
+          GST_VIDEO_FIELD_ORDER_TOP_FIELD_FIRST;
+    else
+      GST_VIDEO_INFO_FIELD_ORDER (out_info) =
+          GST_VIDEO_FIELD_ORDER_BOTTOM_FIELD_FIRST;
   }
 
   if (!gst_structure_has_field (in_s, "chroma-site")) {
@@ -1321,6 +1432,13 @@ gst_ffmpegviddec_negotiate (GstFFMpegVidDec * ffmpegdec,
     output_state->caps = gst_caps_make_writable (output_state->caps);
   }
 
+  if (flags & GST_VIDEO_BUFFER_FLAG_ONEFIELD) {
+    /* TODO: we don't get that information from ffmpeg, so copy it from
+     * the parser */
+    gst_caps_features_add (gst_caps_get_features (ffmpegdec->output_state->caps,
+            0), GST_CAPS_FEATURE_FORMAT_INTERLACED);
+  }
+
   if (!gst_structure_has_field (in_s, "mastering-display-info")) {
     AVFrameSideData *sd = av_frame_get_side_data (picture,
         AV_FRAME_DATA_MASTERING_DISPLAY_METADATA);
@@ -1656,7 +1774,7 @@ gst_ffmpegviddec_video_frame (GstFFMpegVidDec * ffmpegdec,
       ! !(ffmpegdec->picture->flags & AV_FRAME_FLAG_CORRUPT));
 
   if (!gst_ffmpegviddec_negotiate (ffmpegdec, ffmpegdec->context,
-          ffmpegdec->picture))
+          ffmpegdec->picture, GST_BUFFER_FLAGS (out_frame->input_buffer)))
     goto negotiation_error;
 
   pool = gst_video_decoder_get_buffer_pool (GST_VIDEO_DECODER (ffmpegdec));
@@ -1704,7 +1822,7 @@ gst_ffmpegviddec_video_frame (GstFFMpegVidDec * ffmpegdec,
     if (side_data) {
       GST_LOG_OBJECT (ffmpegdec,
           "Found CC side data of type AV_FRAME_DATA_A53_CC, size %d",
-          side_data->size);
+          (int) side_data->size);
       GST_MEMDUMP ("A53 CC", side_data->data, side_data->size);
 
       /* do not add closed caption meta if it already exists */
@@ -1747,6 +1865,8 @@ gst_ffmpegviddec_video_frame (GstFFMpegVidDec * ffmpegdec,
             GST_TIME_FORMAT, tmp, tmp->system_frame_number,
             GST_TIME_ARGS (tmp->pts), GST_TIME_ARGS (tmp->dts));
         /* drop extra ref and remove from frame list */
+        GST_VIDEO_CODEC_FRAME_FLAG_UNSET (tmp,
+            GST_FFMPEG_VIDEO_CODEC_FRAME_FLAG_ALLOCATED);
         gst_video_decoder_release_frame (dec, tmp);
       } else {
         /* drop extra ref we got */
@@ -1759,9 +1879,26 @@ gst_ffmpegviddec_video_frame (GstFFMpegVidDec * ffmpegdec,
 
   av_frame_unref (ffmpegdec->picture);
 
+  if (frame)
+    GST_VIDEO_CODEC_FRAME_FLAG_UNSET (frame,
+        GST_FFMPEG_VIDEO_CODEC_FRAME_FLAG_ALLOCATED);
+
+  if (gst_video_decoder_get_subframe_mode (GST_VIDEO_DECODER (ffmpegdec)))
+    gst_video_decoder_have_last_subframe (GST_VIDEO_DECODER (ffmpegdec),
+        out_frame);
+
   /* FIXME: Ideally we would remap the buffer read-only now before pushing but
    * libav might still have a reference to it!
    */
+  if (GST_BUFFER_FLAG_IS_SET (out_frame->input_buffer,
+          GST_VIDEO_BUFFER_FLAG_ONEFIELD)) {
+    GST_BUFFER_FLAG_SET (out_frame->output_buffer,
+        GST_VIDEO_BUFFER_FLAG_ONEFIELD);
+    if (GST_BUFFER_FLAG_IS_SET (out_frame->input_buffer,
+            GST_VIDEO_BUFFER_FLAG_TFF)) {
+      GST_BUFFER_FLAG_SET (out_frame->output_buffer, GST_VIDEO_BUFFER_FLAG_TFF);
+    }
+  }
   *ret =
       gst_video_decoder_finish_frame (GST_VIDEO_DECODER (ffmpegdec), out_frame);
 
@@ -1774,12 +1911,15 @@ beach:
 no_output:
   {
     GST_DEBUG_OBJECT (ffmpegdec, "no output buffer");
+    GST_VIDEO_CODEC_FRAME_FLAG_UNSET (frame,
+        GST_FFMPEG_VIDEO_CODEC_FRAME_FLAG_ALLOCATED);
     gst_video_decoder_drop_frame (GST_VIDEO_DECODER (ffmpegdec), out_frame);
     goto beach;
   }
 
 negotiation_error:
   {
+    gst_video_decoder_drop_frame (GST_VIDEO_DECODER (ffmpegdec), out_frame);
     if (GST_PAD_IS_FLUSHING (GST_VIDEO_DECODER_SRC_PAD (ffmpegdec))) {
       *ret = GST_FLOW_FLUSHING;
       goto beach;
@@ -1876,8 +2016,10 @@ gst_ffmpegviddec_handle_frame (GstVideoDecoder * decoder,
   }
 
   /* treat frame as void until a buffer is requested for it */
-  GST_VIDEO_CODEC_FRAME_FLAG_SET (frame,
-      GST_VIDEO_CODEC_FRAME_FLAG_DECODE_ONLY);
+  if (!GST_VIDEO_CODEC_FRAME_FLAG_IS_SET (frame,
+          GST_FFMPEG_VIDEO_CODEC_FRAME_FLAG_ALLOCATED))
+    GST_VIDEO_CODEC_FRAME_FLAG_SET (frame,
+        GST_VIDEO_CODEC_FRAME_FLAG_DECODE_ONLY);
 
   data = minfo.data;
   size = minfo.size;
@@ -1966,8 +2108,9 @@ gst_ffmpegviddec_start (GstVideoDecoder * decoder)
   oclass = (GstFFMpegVidDecClass *) (G_OBJECT_GET_CLASS (ffmpegdec));
 
   GST_OBJECT_LOCK (ffmpegdec);
-  gst_ffmpeg_avcodec_close (ffmpegdec->context);
-  if (avcodec_get_context_defaults3 (ffmpegdec->context, oclass->in_plugin) < 0) {
+  avcodec_free_context (&ffmpegdec->context);
+  ffmpegdec->context = avcodec_alloc_context3 (oclass->in_plugin);
+  if (ffmpegdec->context == NULL) {
     GST_DEBUG_OBJECT (ffmpegdec, "Failed to set context defaults");
     GST_OBJECT_UNLOCK (ffmpegdec);
     return FALSE;
@@ -2234,10 +2377,11 @@ gst_ffmpegviddec_set_property (GObject * object,
     case PROP_DIRECT_RENDERING:
       ffmpegdec->direct_rendering = g_value_get_boolean (value);
       break;
+#ifndef GST_REMOVE_DEPRECATED
     case PROP_DEBUG_MV:
-      ffmpegdec->debug_mv = ffmpegdec->context->debug_mv =
-          g_value_get_boolean (value);
+      /* non-functional */
       break;
+#endif
     case PROP_MAX_THREADS:
       ffmpegdec->max_threads = g_value_get_int (value);
       break;
@@ -2261,17 +2405,19 @@ gst_ffmpegviddec_get_property (GObject * object,
 
   switch (prop_id) {
     case PROP_LOWRES:
-      g_value_set_enum (value, ffmpegdec->context->lowres);
+      g_value_set_enum (value, ffmpegdec->lowres);
       break;
     case PROP_SKIPFRAME:
-      g_value_set_enum (value, ffmpegdec->context->skip_frame);
+      g_value_set_enum (value, ffmpegdec->skip_frame);
       break;
     case PROP_DIRECT_RENDERING:
       g_value_set_boolean (value, ffmpegdec->direct_rendering);
       break;
+#ifndef GST_REMOVE_DEPRECATED
     case PROP_DEBUG_MV:
-      g_value_set_boolean (value, ffmpegdec->context->debug_mv);
+      g_value_set_boolean (value, FALSE);
       break;
+#endif
     case PROP_MAX_THREADS:
       g_value_set_int (value, ffmpegdec->max_threads);
       break;
@@ -2436,8 +2582,7 @@ gst_ffmpegviddec_register (GstPlugin * plugin)
 
     /* (Ronald) MPEG-4 gets a higher priority because it has been well-
      * tested and by far outperforms divxdec/xviddec - so we prefer it.
-     * msmpeg4v3 same, as it outperforms divxdec for divx3 playback.
-     * VC1/WMV3 are not working and thus unpreferred for now. */
+     * msmpeg4v3 same, as it outperforms divxdec for divx3 playback. */
     switch (in_plugin->id) {
       case AV_CODEC_ID_MPEG1VIDEO:
       case AV_CODEC_ID_MPEG2VIDEO:
